{$B-}
unit vartable;
{$DEFINE delphi6}
interface
{$IFDEF delphi6}uses Variants;{$ENDIF}

type
  TKeyStr=shortstring;
  TVarTable=class(TObject)
  protected
    items:variant;
    cacheKey:TKeyStr;
    cacheIndex:longint;
    function  KeyIndex(key:TKeyStr):longint;
    function  Key(const index:longint):TKeyStr;
    procedure AddArrayItem(var v:variant; const index,value:variant);
  public
    function  Item(const key:TKeyStr):variant;
    function  Assign(const key:TKeyStr; const value:variant; const index:variant):integer; overload;
    function  Assign(const key:TKeyStr; const value:variant):integer; overload;
    function  Add(key:TKeyStr; const value:variant; const typ:smallint=0):integer;
    function  AssignOrAdd(const key:TKeyStr; const value:variant):integer;
    function  ArrayIndex(const key:TKeyStr; const index:variant):integer; overload;
    function  ArrayIndex(const key:integer; const index:variant):integer; overload;
    function  ArrayItem(const key:TKeyStr; const index:variant):variant; overload;
    function  ArrayItem(const key:integer; const index:variant):variant; overload;
    function  Swap(i1,i2:longint):boolean;
    function  Remove(const key:TKeyStr):integer;
    function  Exists(const key:TKeyStr):boolean;
    function  StripTo(const key:TKeyStr):integer;
    function  GetAttr(const key:TKeyStr):integer;
    function  SetAttr(const key:TKeyStr; const attr:integer):integer;
  end;
  
const
  // indexy items[indexpromenne][]
  vtKEY     =0;
  vtVALUE   =1;
  vtATTR    =2;
  vtTYPE    =3;
  // bity v [vtATTR]
  vtVARIABLE=0;
  vtCONSTANT=1;
  vtFUNCTION=2;

function VarCompare(const a,b:variant):TVariantRelationship;

implementation

uses SysUtils;

function VarCompare(const a,b:variant):TVariantRelationship;
var
  a_,b_:double;
  a_num,b_num:boolean;
begin
  a_num:=VarIsNumeric(a); b_num:=VarIsNumeric(b);
  if a_num xor b_num then
    if a_num then result:=vrLessThan else result:=vrGreaterThan
  else
    if a_num and b_num then begin
      a_:=VarAsType(a,VarDouble); b_:=VarAsType(b,VarDouble);
      if a_<b_ then result:=vrLessThan
      else if a_>b_ then result:=vrGreaterThan
      else result:=vrEqual;
    end else
      if a<b then result:=vrLessThan
      else if a>b then result:=vrGreaterThan
      else result:=vrEqual;
end;

Function TVarTable.KeyIndex(key:TKeyStr):longint;
var i:longint;
begin
  key:=LowerCase(key);
  if (key=cacheKey) and (cacheKey<>'') then begin
    result:=cacheIndex;
    exit;
  end;
  result:=-1;
  if not VarIsArray(items) then exit;
  i:=VarArrayHighBound(items,1);
  while i>=0 do begin
    if items[i][vtATTR] and vtTYPE=vtFUNCTION then
      if (key='result') or (key=items[i][vtKEY]) then begin
        result:=i;
        break;
      end else begin
        i:=0;
        while i<=VarArrayHighBound(items,1) do begin
          if items[i][vtATTR] and vtTYPE=vtFUNCTION then
            break
          else if items[i][vtKEY]=key then begin
            result:=i;
            i:=0;
            break;
          end;
          inc(i);
        end;
      end
    else if items[i][vtKEY]=key then begin
      result:=i;
      if key<>'result' then begin
        cacheKey:=key;
        cacheIndex:=result;
      end;
      break;
    end;
    dec(i);
  end;
end;  {.KeyIndex}

Function TVarTable.Key(const index:longint):TKeyStr;
begin
  if (index<0) or (index>VarArrayHighBound(items,1)) then result:=''
  else result:=items[index][vtKEY];
end;

Function TVarTable.Item(const key:TKeyStr):variant;
var i:longint;
begin
  result:=null;
  i:=KeyIndex(key);
  if i>=0 then result:=items[i][vtVALUE];
end;

Function TVarTable.Add(key:TKeyStr; const value:variant; const typ:smallint=0):integer;
begin
  key:=LowerCase(key);
  if VarIsArray(items) then
    VarArrayRedim(items,VarArrayHighBound(items,1)+1)
  else
    items:=VarArrayCreate([0,0],varVariant);
  cacheKey:=key;
  cacheIndex:=VarArrayHighBound(items,1);
  items[cacheIndex]:=VarArrayOf([key,value,typ]);
  result:=cacheIndex;
end;

Function TVarTable.Swap(i1,i2:longint):boolean;
var u,v:variant;
begin
  result:=false;
  v:=VarArrayHighBound(items,1);
  if (i1=i2) or (i1<0) or (i2<0) or (i1>v) or (i2>v) then exit;
  if (items[i1][vtATTR] and 1<>0) or (items[i2][vtATTR] and 1<>0) then exit;
  u:=items[i1];
  v:=items[i2];
  items[i1]:=v;
  items[i2]:=u;
  items[i1]:=u;
  items[i2]:=v;
  result:=true;
  cacheKey:='';
end;

Procedure TVarTable.AddArrayItem(var v:variant; const index,value:variant);
var i:integer;
begin
  if VarIsArray(v) then i:=VarArrayHighBound(v,1) else i:=-1;
  if i>=0 then VarArrayRedim(v,i+2) else v:=VarArrayCreate([0,1],varVariant);
  v[succ(i)]:=index;
  v[succ(succ(i))]:=value;
end;

Function TVarTable.Assign(const key:TKeyStr; const value:variant; const index:variant):integer; // index - prirazujeme-li do normalni promenne, predame Null. Je-li promenna pole, predame jeho index. Prirazeni pak neovlivni celou promennou, ale jen prvek s timto indexem.
var t,u:variant; i:integer;
begin
  result:=KeyIndex(key);
  if result=-1 then exit;
  if items[result][vtATTR] and vtCONSTANT<>0 then
    result:=-1 // jedna se o konstantu - vratime chybovou navratovou hodnotu
  else begin
    t:=items[result];
    if index<>Null then begin // promenna je pole - priradime hodnotu do jejiho prvku
      i:=ArrayIndex(result,index);
      u:=t[vtVALUE];
      if i=-1 then AddArrayItem(u,index,value) // index nebyl nalezen - vytvorime jej
      else u[succ(i)]:=value; // index pole je znamy - priradime do nej
      t[vtVALUE]:=u;
    end else t[vtVALUE]:=value; // promenna neni pole
    items[result]:=t;
  end;
end;

Function TVarTable.Assign(const key:TKeyStr; const value:variant):integer;
begin
  result:=Assign(key,value,Null);
end;

Function TVarTable.AssignOrAdd(const key:TKeyStr; const value:variant):integer;
begin
  result:=Assign(key,value);
  if result=-1 then result:=Add(key,value);
end;

Function  TVarTable.ArrayIndex(const key:integer; const index:variant):integer;
var i:integer;
begin
  result:=-1;
  if not VarIsArray(items) then exit else
    if (key<0) or (key>VarArrayHighBound(items,1)) then exit;
  if not VarIsArray(items[key][vtVALUE]) then exit;
  i:=pred(VarArrayHighBound(items[key][vtVALUE],1));
  while i>0 do
    if VarCompare(items[key][vtVALUE][i],index)=vrEqual then begin
      result:=i;
      break;
    end else dec(i,2);
end;

Function  TVarTable.ArrayIndex(const key:TKeyStr; const index:variant):integer;
begin
  result:=KeyIndex(key);
  if result=-1 then exit
  else result:=ArrayIndex(result,index);
end;

Function  TVarTable.ArrayItem(const key:integer; const index:variant):variant;
var i:integer;
begin
  result:=-1;
  if not VarIsArray(items) then exit else
    if (key<0) or (key>VarArrayHighBound(items,1)) then exit;
  if not VarIsArray(items[key][vtKEY]) then exit;
  i:=pred(VarArrayHighBound(items[key][vtKEY],1));
  while i>0 do
    if VarCompare(items[i][vtKEY],index)=vrEqual then begin
      result:=items[i][vtVALUE];
      break;
    end else dec(i,2);
end;

Function  TVarTable.ArrayItem(const key:TKeyStr; const index:variant):variant;
begin
  result:=KeyIndex(key);
  if result=-1 then exit
  else result:=ArrayItem(result,index);
end;

Function TVarTable.Remove(const key:TKeyStr):integer;
var m:longint;
begin
  result:=KeyIndex(key);
  if result<0 then exit;
  m:=VarArrayHighBound(items,1);
  while result<m do begin
    items[result]:=items[result+1];
    inc(result);
  end;
  VarArrayRedim(items,m-1);
  cacheKey:='';
end;

Function TVarTable.Exists(const key:TKeyStr):boolean;
begin
  result:=KeyIndex(key)>=0;
end;

Function TVarTable.StripTo(const key:TKeyStr):integer;
begin
  result:=KeyIndex(key);
  if result<0 then exit;
  VarArrayRedim(items,pred(result));
  cacheKey:='';
end;

Function TVarTable.GetAttr(const key:TKeyStr):integer;
begin
  result:=KeyIndex(key);
  if result<0 then result:=Null else result:=items[result][vtATTR];
end;

Function TVarTable.SetAttr(const key:TKeyStr; const attr:integer):integer;
var t:variant;
begin
  result:=KeyIndex(key);
  if result<0 then exit;
  t:=items[result];
  t[vtATTR]:=attr;
  items[result]:=t;
end;

end.

